/* Copyright (c) 1996 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     Config

   PURPOSE
     define configuration and op's on it

   NOTES
     
   AUTHOR
     David Hsu  
   
     dyhsu@cs.stanford.edu
     http://robotics.stanford.edu/~dyhsu 
                  
   HISTORY
     created July, 1996

***************************************************************************/


#ifndef _CONFIG_H
#define _CONFIG_H

#include "math++.h"
#include "utilities.H"


#define DIM 	7			/* dim of c-space */


class Config;				/* forward decl */

/* Coords
 *    class for coord's of c-space
 */

class Coords {
    friend istream& operator>> (istream& , Coords& );
    friend ostream& operator<< (ostream& , Coords& );
    friend Bool inBox(Config& q0, Config& q1, Coords& eps, Coord a);

public:
    Coord& operator[] (int i) { return q[i]; }
    Coord* coords() {return q;}

protected:
    Coord q[DIM];
};


/* Config
 *    class for representation of configuration
 *    q[0],q[1],q[2] gives translation in 3D
 *    q[3],q[4],q[5],q[6] gives rotation in 3d in quaternion representation 
 *
 *    Invariant: The quaternion part is stored as a unit quaternion
 */

class Config:public Coords {
    friend Coord wdist(Config& p, Config& q, Coord param);
    friend Bool inBox(Config& q0, Config& q1, Coords& eps, Coord a);

public:
    Coord dist;				/* dist to obstacle */
    Config* link;    			/* link with other config's to
					   represent connectivity of c-space */
    int  wgt;				/* weight of the node */

    void trans(Coord x, Coord y, Coord z) {q[0]=x; q[1]=y; q[2]=z;}
    void rot(Coord r0, Coord r1, Coord r2, Coord r3) {
	q[3]=r0; q[4]=r1; q[5]=r2; q[6]=r3; 
	normalize();
    }
    inline void midpoint(Config& q0, Config& q1);
    inline void normalize();
    inline void antipode(Config& cfg) ;
    inline void frame( Frame3& fs);
};

/* midpoint
 *    set the coords to be the midpoint of q0 and q1
 */

inline void Config::midpoint(Config& q0, Config& q1)
{
    for (int i=0; i < DIM; i++)
	q[i] = (q1.q[i] + q0.q[i])/ 2.0;
    normalize();
}

    
/* normalize
 *    make the quaternion part of unit length
 */

inline void Config::normalize() 
{
    Coord d = sqrt_pp( q[3]*q[3] + q[4]*q[4] + q[5]*q[5] + q[6]*q[6] );
    q[3] /= d; q[4] /= d; q[5] /= d; q[6] /= d; 
}


/* antipode
 *     move the quaternion onto the same hemisphere as that of cfg by
 *     applying antipodal map
 */

inline void Config::antipode(Config& p) 
{
    Coord d0,d1;

    d0 = ( q[3] - p.q[3] ) * ( q[3] - p.q[3] ) +
   	 ( q[4] - p.q[4] ) * ( q[4] - p.q[4] ) +
	 ( q[5] - p.q[5] ) * ( q[5] - p.q[5] ) +
	 ( q[6] - p.q[6] ) * ( q[6] - p.q[6] );
    d1 = ( q[3] + p.q[3] ) * ( q[3] + p.q[3] ) +
	 ( q[4] + p.q[4] ) * ( q[4] + p.q[4] ) +
	 ( q[5] + p.q[5] ) * ( q[5] + p.q[5] ) +
	 ( q[6] + p.q[6] ) * ( q[6] + p.q[6] );
    if (d0 > d1) {
	q[3] = -q[3]; q[4] = -q[4]; q[5] = -q[5]; q[6] = -q[6]; 
    }
}


/* frame
 *  convert the representation in C-space to Cartesian space
 *  q is represented as (x,y,z, q0, q1, q2, q3) where (x,y,z) specifies
 *  position and (q0, (q1,q2,q3)) is a unit quoternion 
 *  fs is represented as (x,y,z) and a 3x3 rotation matrix 
 */
inline void Config::frame( Frame3& fs) 
{
    fs.p.x = q[0]; fs.p.y = q[1]; fs.p.z = q[2];

    fs.r[0][0] = 2*( q[3]*q[3] + q[4]*q[4] )-1;
    fs.r[0][1] = 2*( q[4]*q[5] - q[3]*q[6] );
    fs.r[0][2] = 2*( q[4]*q[6] + q[3]*q[5] );

    fs.r[1][0] = 2*( q[4]*q[5] + q[3]*q[6] );
    fs.r[1][1] = 2*( q[3]*q[3] + q[5]*q[5] )-1;
    fs.r[1][2] = 2*( q[5]*q[6] - q[3]*q[4] );

    fs.r[2][0] = 2*( q[4]*q[6] - q[3]*q[5] );
    fs.r[2][1] = 2*( q[5]*q[6] + q[3]*q[4] );
    fs.r[2][2] = 2*( q[3]*q[3] + q[6]*q[6] )-1;
}





/* dist
 *   compute the distance between two configurations in workspace
 *   
 *   param: a parameter used in the definition of the distance
 *          It may or may not need to be used depending on the definition
 *          of distance. Here param is the max radius of the objects.
 */

inline Coord wdist(Config& p, Config& q, Coord param)
{
   Coord d, quat[4], theta;

   d  = sqrt_pp(( q.q[0] - p.q[0] ) * ( q.q[0] - p.q[0] ) +
		( q.q[1] - p.q[1] ) * ( q.q[1] - p.q[1] ) +
		( q.q[2] - p.q[2] ) * ( q.q[2] - p.q[2] ) ); 
   QuatMulConj( quat, p.q+3, q.q+3);		/* compute the rotation */
   theta  =  acos_pp( ABS( quat[0]) ) * 2;      /* compute the rotation angle*/
   d  += theta * param;
   return d;
}


/* inBox
 *    test whether two config's are close enough
 *    true if  a * | q0[i] - q1[i] | < eps[i] for all i
 *    
 */

inline Bool inBox(Config& q0, Config& q1, Coords& eps, Coord a=1.0)
{
    q0.antipode(q1);

    Bool result = TRUE;
    for (int i=0; i<3 && result ; i++) 
	result =  ( ABS(q0.q[i] - q1.q[i])*a) <= eps.q[i] ;

    /* watch out multiplying by 2 
    for (; i<DIM && result ; i++) 
	result =  ( ABS(q0.q[i] - q1.q[i])*a) <= 4 ;  */
    return result;
}


inline  Coord dot(Coord* q0, Coord* q1) 
{
    return ( q0[0] * q1[0] + q0[1] * q1[1] + q0[2] * q1[2] + 
             q0[3] * q1[3] + q0[4] * q1[4] + q0[5] * q1[5] + 
	     q0[6] * q1[6] );
}

inline Coord dotTrans(Coord* q0, Coord* q1)
{
    return ( q0[0] * q1[0] + q0[1] * q1[1] + q0[2] * q1[2] );
}

inline Coord dotRot(Coord* q0, Coord* q1)
{
    return ( q0[3] * q1[3] + q0[4] * q1[4] + q0[5] * q1[5] + 
	     q0[6] * q1[6] );
}



typedef Config* PtrConfig;

class ConfigArray;			/* forward decl */
class Cspace;			
class Robot;

ConfigArray* interpolate(Config& p, Config& q, Robot& robo);
    
#endif /* _CONFIG_H */
