/* Copyright (c) 1996 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     Cspace

   PURPOSE
     define c-space and sample points from c-space
     
   NOTES
     
   AUTHOR
     David Hsu  
   
     dyhsu@cs.stanford.edu
     http://robotics.stanford.edu/~dyhsu 
                  
   HISTORY
     created July, 1996

***************************************************************************/


/* sample
 *   generate a random sample from the c-space
 *
 *   origin : center of the sample box
 *   box    : size of sample box
 *   s      : return value 
 */

inline Bool Cspace::sample(Config& origin,SampleBox&  box, Config& s)
{
    Coord *cq = origin.coords();	/* coords of center of box */
    Coord *sq = s.coords();		/* coords of sample point  */
    Coord *b = box.box();
    const double *nq;
    float radius2, rq[DIM];
    Bool loop = TRUE;

    while (loop) {
      nq = randDbl.get(DIM); Stats.totalRawSample++;

      // translation part
      for (int i=0; i < 3; i++) {
	sq[i] = cq[i] + ( nq[i] - 0.5 ) * b[i];
      }

      // return false if sample out of the boundary
      if ( sq[0] < l[0] || sq[0] > r[0]  ||
	   sq[1] < l[1] || sq[1] > r[1]  ||
	   sq[2] < l[2] || sq[2] > r[2] )
	return FALSE;

      // rotation part
      radius2 = 0.0;
      for (i=3 ; i<DIM; i++  ) {
	rq[i] = nq[i] - 0.5;
	radius2 += rq[i] * rq[i];
      }

      // reject points outside  sphere of radius 0.5
      if (radius2 > 0.25)  continue;
      for (i=3; i < DIM; i++)
	sq[i] = cq[i] + rq[i]*b[i];

      // normalize the quaternion
      s.normalize();
      loop = FALSE;
    }
    return TRUE;
}

/* sample
 *   generate a random sample from the c-space
 *
 *   origin : center of the sample box
 *   s      : return value 
 *   param  : max radius of robot
 */

inline Bool Cspace::sample(Config& origin, Config& s, float param)
{
    Coord *cq = origin.coords();	/* coords of center of box */
    Coord *sq = s.coords();		/* coords of sample point  */
    Coord b[DIM];			/* sampling box */
    const double *nq;
    float radius2, rq[DIM];
    Bool loop = TRUE;

    Coord d = MIN(Ct.boxSize, MAX(Ct.boxSize , origin.dist * 15) );

    // compute max movement in translation and rotation
    for (int i=0; i<3; i++)
        b[i] = MIN(d, r[i] - l[i]);
    for (; i<DIM; i++) 
        b[i] = MIN( 2 * d / param, r[i] - l[i]) ;

    while (loop) {
      nq = randDbl.get(DIM); Stats.totalRawSample++;

      // translation part
      for (int i=0; i < 3; i++) {
	sq[i] = cq[i] + ( nq[i] - 0.5 ) * b[i];
      }

      // return false if sample out of the boundary
      if ( sq[0] < l[0] || sq[0] > r[0]  ||
	   sq[1] < l[1] || sq[1] > r[1]  ||
	   sq[2] < l[2] || sq[2] > r[2] )
	return FALSE;

      // rotation part
      radius2 = 0.0;
      for (i=3 ; i<DIM; i++  ) {
	rq[i] = nq[i] - 0.5;
	radius2 += rq[i] * rq[i];
      }

      // reject points outside  sphere of radius 0.5
      if (radius2 > 0.25)  continue;

      for (i=3; i < DIM; i++)
	sq[i] = cq[i] + rq[i]*b[i];

      // normalize the quaternion
      s.normalize();
      loop = FALSE;
    }
    return TRUE;
}





/* sample
 *   generate a random direction from the c-space
 *
 *   dir : return value
 */

inline static void Cspace::sampleDir(Coords& dir)
{
    Coord *dq = dir.coords();
    const double *rq = randDbl.get(DIM);

    for (int i=0;i < DIM; i++)
	dq[i] = rq[i];
}


