/* Copyright (c) 1996 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     Model

   PURPOSE
     data structures for storing models

   NOTES
     
   AUTHOR
     David Hsu  
   
     dyhsu@cs.stanford.edu
     http://robotics.stanford.edu/~dyhsu 
                  
   HISTORY
     created July, 1996

***************************************************************************/


#ifndef _MODEL_H
#define _MODEL_H

#include "dist.h"
#include "utilities.H"

/* ModelType
 *
 */

typedef enum { P_TYPE=112, C_TYPE=99, S_TYPE=115 } ModelType;


/* Model
 *    abstract base class providing a template for different models
 */

class Model {
    friend ostream& operator<< (ostream& os, Model& m);
public:
    Model() { maxRadius = 0; handle = NULL; }
    virtual ~Model() {} ;

    virtual void buildModel()=0; 	/* build model for collision checker */
    Sobj3* model() { return handle;}
    virtual void  calcRadius()=0;
    Coord radius() { return maxRadius; }

protected:
    enum ModelType  type;		/* type of model  */
    Coord maxRadius;                    /* max radius of the model */
    Sobj3 *handle;                      /* data structure for collision
                                           checker */
    virtual void print(ostream& os)=0;
};



/* Triangle 
 *   type for a face of a polyhedron
 */

typedef int Tri[3];



/* Phdn
 *   class for polyhedral model
 */

class Phdn: public Model {
    friend istream& operator>> (istream& is, Phdn& p);
    friend ostream& operator<< (ostream& os, Phdn& p);

public:
    Phdn() { v=NULL; t=NULL; vcount=tcount=0; type=P_TYPE; tol=0.0;}
    ~Phdn() { delete [] v; delete [] t; }

    void bgn(int vc, int tc) { v=new Vec3[vcount=vc]; t=new Tri[tcount=tc]; }
    void end();

private:
    int vcount;				/* number of vertices */
    int tcount;				/* number of triangle */
    Vec3 *v;				/* vertices */
    Tri *t;				/* triangles */
    Coord tol;				/* tolerance for acccomodating 
					   inprecision in data */

    void calcRadius();
    void buildModel();
    void print(ostream& os);
};


class Moray {
    friend istream& operator>> ( istream& is, Moray& m);
    friend ostream& operator<< ( ostream& os, Moray& m);

public:
    Moray() { m = NULL; mcount=0; maxRadius=0.;handle=NULL; }
    virtual ~Moray();

    void bgn(int mc) { m = new Model*[mcount=mc]; }
    void end();
    void model(int ind, Model *model) { m[ind] = model;}
    Sobj3* model() {return handle;}
    Coord radius() { return maxRadius; }
    
protected:
    int mcount;				/* number of models */
    Model** m;				/* models */
    Coord maxRadius;			/* max radius of all models */
    Sobj3 *handle;		        /* data structure for collsion 
					   checker*/
    void calcRadius();
    virtual void buildModel();
};

#endif  /* _MODEL_H */
