/* Copyright (c) 1996 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     Planner1

   PURPOSE
     planning algo.

   NOTES
     
   AUTHOR
     David Hsu  
   
     dyhsu@cs.stanford.edu
     http://robotics.stanford.edu/~dyhsu 
                  
   HISTORY
     created July, 1996

***************************************************************************/


#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <sys/types.h>
#include <sys/times.h>
#include "config.H"
#include "global.H"
#include "planner1.H"
#include "cspace.il.H"

/*------------------------- class Tray -------------------------*/

void Tray::bgn( Robot* robot, Obstacle* obstacle, Cspace* cspace, 
	        SampleBox* samplebox)
{
  r = robot;				
  obs = obstacle;
  csp = cspace;
  box = samplebox;
}


/* reset
 *   reset the states of Tray
 */

Bool Tray::reset(Config& n)
{
  dead = white = gray = black = 0;
  *qs[0] = n;				/* root */
  qs[0]->link = NULL; qs[0]->wgt = 1;

  // check if the root is in collision
  if ( r->clearance( *qs[0] ) > Ct.tol  )   {
    white++;
    n.dist = qs[0]->dist;
    return TRUE;
  }
  else {
    // guess a free config close-by
    Bool free;

    cerr << "finding free start point\n" ;
    for (int i=0; i < 300 && !(free = r->clearance( *qs[0] ) <= Ct.tol );i++)
      csp->sample(n, *qs[0], r->radius());
    if (free) n = *qs[0];
    return FALSE;
  }
}

void Tray::expand()
{
  int pid;
  Bool loop1,  loop2 ;
  const double *rw;
  const long *cn;
  const int size = 10;

//   // select nodes as center of expansion
//   cn = randInt.get(Ct.expNum);

//   for (int i=0; i < Ct.expNum; i++) {

//     // if there are white nodes, select one
//     if ( white > gray ) {
//       pid = cn[i] % (white - gray) + gray;    
// #ifdef DEBUG
//       cerr << "\nS: select white node " << pid << endl;
// #endif
            
//       // swap the selected node and the first white node, which has
//       // index gray
//       swap( qs[gray], qs[pid] );
//       pid = gray;
            
//       // increment gray counter to mark the selected node gray
//       gray++;
//     }
        
//     // otherwise if there are gray nodes, select one
//     else if ( gray > dead ) {
//       pid = cn[i] % (gray - dead) + dead;
// #ifdef DEBUG
//       cerr << "\nS: select gray node " << pid << endl;
// #endif
//     }

//     // otherwise select a dead node
//     else if ( dead > 1 ) {
//       pid = cn[i] % (dead - 1) + 1;
// #ifdef DEBUG
//       cerr << "\nS: select dead node " << pid << endl;
// #endif
//     }
//     else {
//       pid = 0;
// #ifdef DEBUG
//       cerr << "S: select root node " << pid << endl;
// #endif
//     }

//     if ( ! sample( *qs[pid] ) ) {
//       if ( pid >= dead ) {
//         // swap the node with the first gray node
//         swap( qs[dead], qs[pid] );

//         // mark the node dead
//         dead++;
// #ifdef DEBUG
//         cerr << "K: kill dead node --- " << endl;
// #endif
//       }
//       else {
// #ifdef DEBUG
//         cerr << "W: wasting time on dead nodes --- " << endl;
// #endif
//       }
//     }
//   }


  for (int i=0; i < Ct.expNum; i++) {
    loop1 = loop2 = TRUE;
    
    while (loop1) {
      // select nodes as center of expansion
      cn = randInt.get(size);
      rw = randDbl.get(size);
      for (int j=0; j < size && loop2; j++) {
	pid = cn[j] % white;
	if ( rw[j] <= 1.0/ qs[pid]->wgt) {
	  loop1=loop2=FALSE;
	}
      }
    }
#ifdef DEBUG
    cerr << "S: select node " << pid << " weight " << qs[pid]->wgt 
	 << " dist " << qs[pid]->dist << endl;
    cerr << "\t" << (*qs[pid])<< endl;
#endif

    sample( *qs[pid] );
  }
}


const static int NUM_TEST = 8;
const static int MAX_NUM_NBD = 100;

Bool Tray::getChild(Config& parent, Config& child)
{
  int i,j, count;
  int nbd[MAX_NUM_NBD];

  // get a new sample point from c-free space
  if (! (csp->sample( parent, *box, child )) )
    return FALSE;
    
  // make the sampling density uniform by rejection
  // give higher probability to those config's on the convex hull locall
  count = 0;
  for (i=0; i < white && count < MAX_NUM_NBD; i++) {
    if ( inBox( child, *qs[i], *box, 2) ) 
      nbd[count++] = i;
  }
  child.wgt = count + 1;

  // check whether sample is on the convex hull
  int bigger, smaller;
  Bool extreme=FALSE;

  Coords randDir;
  Coord* rq = randDir.coords();
  for (i=0; i < NUM_TEST && !extreme; i++) {

    bigger = smaller = 0;
    Cspace::sampleDir(randDir);
    Coord weight = dot(rq, child.coords());

    for (j=0; j < count; j++) {
      Coord *qq = qs[nbd[j]]->coords();
      if ( dot(rq, qq) > weight )
	bigger++;
      else 
	smaller++;
    }

    // if either the bigger or smaller is 0, the point is extreme
    extreme = ( bigger == 0 || smaller == 0 );
  }   


  if ( (*(randDbl.get(1)) ) > 1.0/count ) {
#ifdef DEBUG
    cerr << "R: "<<  " reject count = " << count << endl;
#endif	
    return FALSE;
  }

  if (extreme) {
#ifdef DEBUG
    cerr << "E: extreme point found " << " count=" << count << endl;
#endif
  }
  else if ( (*(randDbl.get(1))) > 1/4.0) {
    cerr << "K: reject non extreme samples " << "count = " << count << endl ;
    return FALSE;
  }


  Stats.totalSample++;
  // check whether the config is in c-free
  if ( r->clearance(child) > Ct.tol  ) {

    Stats.totalFree++; Stats.numLink++;
    if ( r->connectPair( parent, child ) ) {
      child.link = &parent; 
      Stats.totalFreeCnt++;
      for (int i=0; i < count;i++) 
        qs[nbd[i]]->wgt++;
      return TRUE;
    }
  }
  return FALSE;
}
    

Bool Tray::sample(Config& parent)
{
  // check array bound
  if ( white + Ct.branchFac > qs.size() )
    qs.resize( qs.size() + 1024);

  for (int j=0; j < Ct.branchFac; j++) {
    Config& child = *qs[white];
    if ( getChild( parent, child ) ) { 
#ifdef DEBUG	    
      cerr << "D: node " << white << " has dist " <<  
	qs[white]->dist  << endl;
      cerr << "   " << (* qs[white] ) << endl;
#endif
      white++;
    }
  }

  return TRUE;
}


void Tray::dump(char* fname)
{
  ofstream target(fname);
  if ( !target )
    error("dump: could not open file ", fname);
  target << (*this);
} 

ostream& operator<<(ostream& os, Tray& t)
{
  for (int i=0; i < t.white ; i++)
    os << (*(t.qs[i])) << endl;
  return os;
}

/*------------------------- class Planner ----------------------*/

void Planner1::bgn(istream& rsrc, istream& osrc, istream &asrc)
{
  BasicPlanner::bgn(rsrc, osrc, asrc);

  // make sample box
  box = new SampleBox(Ct.boxSize, r.radius(), csp );
  cerr << "Box: \n" << (*box) << endl;

  icomp.bgn(&r, &obs, &csp, box);
  gcomp.bgn(&r, &obs, &csp, box);
}


void Planner1::query(ostream& pfile)
{
  char ifname[32], gfname[32];
  PtrConfig iConfig, gConfig;

  found = FALSE;
  if ( !(icomp.reset(init) && gcomp.reset(goal) ) ) {
    cerr << "\n\n *** \x7rss: initial or goal configuration is in C-obstacle ***\n";
    pfile << init << endl << goal << endl;
    return ;
  }
    
  times( & Stats.qStart );
  for (int i=0; i < Ct.MAX_ITER && !found; i++) {
    cerr << "\n\n\n  ***   iteration " << i << "   I Component***\n\n\n";
    icomp.expand();

    if (! Ct.uni) {
      cerr << "\n\n\n  ***   iteration " << i << "   G Component***\n\n\n";
      gcomp.expand();
    }

    cerr << "\n\n\n  ***   iteration " << i << "   Connecting***\n\n\n";
    found = connectComp(iConfig, gConfig);

    if ( Ct.tr && (i % 10 == 0 )) {
      sprintf( ifname, "idump%d.dat", i );
      sprintf( gfname, "gdump%d.dat", i );
      icomp.dump(ifname);
      if (! Ct.uni ) gcomp.dump(gfname);
    }
  }
  times( & Stats.qEnd );

  if (Ct.tr) {
    icomp.dump("idump.dat");
    gcomp.dump("gdump.dat");
  }

  Stats.iWhite = icomp.white; Stats.iGray = icomp.gray;
  Stats.iDead  = icomp.dead;
  Stats.gWhite = gcomp.white; Stats.gGray = gcomp.gray;
  Stats.gDead  = gcomp.dead;
	
  if (found) 
    writePath(iConfig, gConfig, pfile);
}



Bool Planner1::connectComp(Config* &iConfig, Config* &gConfig)
{
  int i, j;
  PtrConfig p,q;

  // connect from init to goal
  for (i=icomp.black; i <  icomp.white; i++) {
    for (j=0; j < gcomp.white ; j++, Stats.numConnection++ ) {
      p = icomp[i]; q = gcomp[j];
      if (wdist(*p, *q,r.radius()) < Ct.cDist && r.connectPair(*p, *q)  ) {
 	iConfig = p; gConfig = q;
 	cerr << "R: " << wdist(*p,*q, r.radius()) << endl;
 	return TRUE;
      }
    }
  }

  // connect from goal to init
  for (i=gcomp.black; i < gcomp.white ; i++ ) {
    for (j=0; j < icomp.black; j++, Stats.numConnection++ ) {
      p = icomp[j]; q = gcomp[i];
      if ( wdist(*p, *q, r.radius()) < Ct.cDist &&  r.connectPair( *p, *q ) ) {
	iConfig = p; gConfig = q;
	cerr << "R: " << wdist(*p,*q, r.radius()) << endl;
	return TRUE;
      }
    }
  }

  // mark those nodes that have already been checked
  icomp.black = icomp.white; gcomp.black = gcomp.white;

  return FALSE;
}
